@@ -6,10 +6,9 @@
  *  This is usually NOT needed, use engine_mpi or engine for real distributed version
  * \author Tianqi Chen
  */
-#define _CRT_SECURE_NO_WARNINGS
-#define _CRT_SECURE_NO_DEPRECATE
 #define NOMINMAX
 
+#include <rabit/base.h>
 #include "rabit/internal/engine.h"
 
 namespace rabit {
