@@ -10,35 +10,33 @@
 #include <xgboost/base.h>
 #include <xgboost/data.h>
 
-#include <algorithm>
 #include <memory>
-#include <limits>
-#include <utility>
-#include <vector>
+#include <string>
 
-#include "simple_csr_source.h"
-#include "../common/group_data.h"
-#include "../common/math.h"
-#include "adapter.h"
 
 namespace xgboost {
 namespace data {
 // Used for single batch data.
 class SimpleDMatrix : public DMatrix {
  public:
-  explicit SimpleDMatrix(std::unique_ptr<DataSource<SparsePage>>&& source)
-      : source_(std::move(source)) {}
-
+  SimpleDMatrix() = default;
   template <typename AdapterT>
   explicit SimpleDMatrix(AdapterT* adapter, float missing, int nthread);
 
+  explicit SimpleDMatrix(dmlc::Stream* in_stream);
+  ~SimpleDMatrix() override = default;
+
+  void SaveToLocalFile(const std::string& fname);
+
   MetaInfo& Info() override;
 
   const MetaInfo& Info() const override;
 
-  float GetColDensity(size_t cidx) override;
+  bool SingleColBlock() const override { return true; }
+  DMatrix* Slice(common::Span<int32_t const> ridxs) override;
 
-  bool SingleColBlock() const override;
+  /*! \brief magic number used to identify SimpleDMatrix binary files */
+  static const int kMagic = 0xffffab01;
 
  private:
   BatchSet<SparsePage> GetRowBatches() override;
@@ -46,12 +44,19 @@ class SimpleDMatrix : public DMatrix {
   BatchSet<SortedCSCPage> GetSortedColumnBatches() override;
   BatchSet<EllpackPage> GetEllpackBatches(const BatchParam& param) override;
 
-  // source data pointer.
-  std::unique_ptr<DataSource<SparsePage>> source_;
-
+  MetaInfo info_;
+  SparsePage sparse_page_;  // Primary storage type
   std::unique_ptr<CSCPage> column_page_;
   std::unique_ptr<SortedCSCPage> sorted_column_page_;
   std::unique_ptr<EllpackPage> ellpack_page_;
+  BatchParam batch_param_;
+
+  bool EllpackExists() const override {
+    return static_cast<bool>(ellpack_page_);
+  }
+  bool SparsePageExists() const override {
+    return true;
+  }
 };
 }  // namespace data
 }  // namespace xgboost
